{== Version information ========================================================
Designed by     :  Krzysztof Bacalski        
Created         : 27.03.97                    
Modified        : 22.05.97                    
Current Version : 1.02                        
=== Description ================================================================
I wasn't satisfied with smart components(that care for aligning of their
children controls) so I decided to write my own. This aligning panel takes care
for all its children-controls what means that during self resizing aligning 
panel moves (\resizes) its children. Its biggest advantage is that it calculates 
their coordinates and stores these data in dynamic array as records of previous
positions that are of the real type. In this way you avoid error caused by
rounding calculated position ratio -> after each resizing of the AligningPanel
your controls will be placed in right positions. In components, I met, this
problem was unsolved and in this way after a few resizings you received you
controls dropped at strange locations!!!
=== Tech info ==================================================================
- New Properties :
--AligningActive   - if true during resizing panel takes care for all its 
                     children; automatically disabled when properties:
                     - MoveHorizontaly/Verticaly
                     - SizePanelsBevels
                     - GlueLabels
                     change their values
--MinHeight        - minimal height of the panel down to which it moves its
                     children
--MinWidth         - minimal width of the panel down to which it moves its
                     children
--MoveHorizontaly  - true -> horizontal children movement enabled 
--MoveVerticaly    - true -> vertical children movement enabled 
--SizePanelsBevels - true -> all children panels and bevels are sized 
                     when being moved
--GlueLabels       - true -> causes the labels with assigned FocusControls
                     to moved like the Focuscontrols themselves
--------------------------------------------------------------------------------
 WARNING : I don't check if during AligningActive you move one of the children
           yourself. So if you want to do something like this I recomend to
           set AligningActive to false first!
--------------------------------------------------------------------------------  
=== Usage ======================================================================
  Set AligningActive property to true to align components proportionaly
according to theirs previous positions and panel shapes' change. By manipulating 
with MoveHorizontaly and MoveVerticaly properties you can handle different 
alignings' kinds. Set MinHeight/width to specify the minimal height/width down 
to which panel should move ist children(zero-means no height/widht restrictions)
Set SizePanelsBevels to true if you want to see all children panels and bevels
being resized during their move. Set GlueLabels to true if you want your labels
to keep the same distance to their FocusControls.
================================================================================
 All suggestions are welcomed. 
 You are free to use this component.
 If you appretiate it let me know . 
 E-mail:  marcin@arcadia.tuniv.szczecin.pl (for Chris)
==================================================================== ChriB'97 =}

unit AlPanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls;

type
  ElementType    = Record
    PrevLeft     : Real;
    PrevTop      : Real;
    PrevWidth    : Real;
    PrevHeight   : Real;
 end;
const
  MaxArraySize   = (65520 div SizeOf(ElementType)); // under a 16-bit environment 
type
  ECannotAssignToActive = class (Exception);
  ETooManyControls      = class (Exception);
  ArrayType      = array[1..MaxArraySize] of ElementType;   
  TAligningPanel = class(TCustomPanel)
  private
    { Private declarations }
    P            : ^ArrayType;
    ArraySize    : LongInt;
    Creating     : Boolean;
    PreviousH    : Integer;
    PreviousW    : Integer;
    FActive      : Boolean;
    FMoveH       : Boolean;
    FMoveV       : Boolean;
    FSizePnlBvl  : Boolean;
    FGlueLbls    : Boolean;
    FMinH        : Integer;
    FMinW        : Integer;
    procedure AllocateArray;
    procedure DeallocateArray;    
    procedure InitArray;    
    function  IsActive : Boolean;
    procedure WMParentNotify(var MSG : Tmessage); message WM_PARENTNOTIFY;
    procedure WMSize        (var MSG : Tmessage); message WM_SIZE;    
    procedure SetFActive    (NewValue : Boolean);
    procedure SetFMoveH     (NewValue : Boolean);
    procedure SetFMoveV     (NewValue : Boolean);
    procedure SetFSizePB    (NewValue : Boolean);
    procedure SetFGlueLb    (NewValue : Boolean);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create ( AOwner: TComponent); override;    
    destructor  Destroy; override;
  published
    { Published declarations }
    property AligningActive   : Boolean read FActive     write SetFActive default False;
    property MinHeight        : Integer read FMinH       write FMinH     default 0;
    property MinWidth         : Integer read FMinW       write FMinW     default 0;
    property MoveHorizontaly  : Boolean read FMoveH      write SetFMoveH  default True;
    property MoveVerticaly    : Boolean read FMoveV      write SetFMoveV  default True;    
    property SizePanelsBevels : Boolean read FSizePnlBvl write SetFSizePB default False;
    property GlueLabels       : Boolean read FGlueLbls   write SetFGlueLb default False;
    property Align;
    property Alignment;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;
    property BorderStyle;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Caption;
    property Color;
    property Ctl3D;
    property Font;
    property Locked;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;    
  end;

procedure Register;

implementation

procedure TAligningPanel.SetFGlueLb (NewValue : Boolean);
begin
 if FGlueLbls <> NewValue then if not IsActive then FGlueLbls := NewValue;
end;

procedure TAligningPanel.WMParentNotify(var MSG: Tmessage);
begin 
 inherited;
 if ArraySize <> ControlCount then AligningActive := false;
end;

function  TAligningPanel.IsActive : Boolean;
begin
 Result := true;
 if FActive then
  raise ECannotAssignToActive.Create ('Cannot assign when AligningActive')
 else
  Result := false; 
end;

procedure TAligningPanel.SetFSizePB (NewValue : Boolean);
begin
 if FSizePnlBvl <> NewValue then if not IsActive then FSizePnlBvl := NewValue;
end;

procedure TAligningPanel.SetFMoveH (NewValue : Boolean);
begin
 if FMoveH <> NewValue then if not IsActive then FMoveH := NewValue;
end;

procedure TAligningPanel.SetFMoveV (NewValue : Boolean);
begin
 if FMoveV <> NewValue then if not IsActive then FMoveV := NewValue;
end;

procedure TAligningPanel.SetFActive(NewValue : Boolean);
begin
 if FActive <> NewValue then
  begin
   if NewValue and(ArraySize > 0)then
    InitArray
   else if NewValue and(ArraySize = 0) then
    begin ArraySize := ControlCount; AllocateArray; end
   else DeallocateArray;
   FActive := NewValue;
  end; 
end;

procedure TAligningPanel.AllocateArray;
begin
 if(ArraySize <= MaxArraySize)and(ArraySize > 0)then
  begin
   P := AllocMem(ArraySize * SizeOf(ElementType));
   InitArray; 
  end
 else if ArraySize > MaxArraySize then
  raise ETooManyControls.create('Too many controls to handle');
end;

procedure TAligningPanel.InitArray;
var i : integer;
begin
 for i := 1 to ArraySize do
  with P^[i] do
   begin 
    PrevLeft   := 0; 
    PrevTop    := 0; 
    PrevWidth  := 0;
    PrevHeight := 0;
   end; 
end;

procedure TAligningPanel.DeallocateArray;
begin
 FreeMem(P, ArraySize * SizeOf(ElementType));
 ArraySize := 0;
end;

procedure TAligningPanel.WMSize(var MSG: Tmessage);
var i        : Integer;
    CurrentH : Integer;
    CurrentW : Integer;
    HRatio   : Real;
    WRatio   : Real;
    Control  : TControl;
begin
 if(not FActive)or(not(FMoveH or FMoveV))then begin inherited; exit; end;
 if Creating then begin PreviousH := Height; PreviousW := Width; Creating := false; end; 
 inherited;
 if(HiWord(MSG.LParam) >= FMinH) then CurrentH := Height else CurrentH := FMinH;
 if(LoWord(MSG.LParam) >= FMinW) then CurrentW := Width  else CurrentW := FMinW;   
 if(ControlCount <> 0)then
  begin
   if(ArraySize = 0)then
    begin
     ArraySize := ControlCount; 
     AllocateArray; 
    end
   else if(ControlCount <> ArraySize)then
    begin 
     DeallocateArray; 
     ArraySize := ControlCount; 
     AllocateArray; 
    end;
   HRatio := (CurrentH / PreviousH);
   WRatio := (CurrentW / PreviousW);
   for i := 1 to ControlCount do
    with P^[i] do
     begin
      Control := Controls[i-1];
      if(FGlueLbls)and(Control is TLabel)then
       if TLabel(Control).FocusControl <> nil then
        begin
         if PrevHeight = 0 then
          PrevHeight := TControl(TLabel(Control).FocusControl).Top - TLabel(Control).Top;
         if PrevWidth  = 0 then 
          PrevWidth  := TControl(TLabel(Control).FocusControl).Left - TLabel(Control).Left;
        end;        
      if FMoveH then
       begin
        if PrevLeft = 0 then PrevLeft := Control.Left;
        PrevLeft        := PrevLeft * WRatio;
        Control.Left    := Round(PrevLeft);
        if(FSizePnlBvl)and((Control is TCustomPanel)or(Control is TBevel))then
         begin
          if PrevWidth = 0 then PrevWidth := Control.Width;
          PrevWidth     := PrevWidth * WRatio;
          Control.Width := Round(PrevWidth);         
         end;                                      
       end; 
      if FMoveV then
       begin
        if PrevTop = 0 then PrevTop := Control.Top;
        PrevTop         := PrevTop * HRatio;
        Control.Top     := Round(PrevTop);
        if(FSizePnlBvl)and((Control is TCustomPanel)or(Control is TBevel))then
         begin
          if PrevHeight = 0 then PrevHeight := Control.Height;
          PrevHeight    := PrevHeight * HRatio;
          Control.Height:= Round(PrevHeight);         
         end;        
       end; 
     end;     
   if(FGlueLbls)then
    for i := 1 to ControlCount do
     if (Controls[i-1] is TLabel)then
      with P^[i] do
       begin
        Control := Controls[i-1];
        if TLabel(Control).FocusControl <> nil then
         begin
          Control.Top  := TControl(TLabel(Control).FocusControl).Top  - Round(PrevHeight);
          Control.Left := TControl(TLabel(Control).FocusControl).Left - Round(PrevWidth);
         end;               
       end;
   PreviousH  := CurrentH;
   PreviousW  := CurrentW;    
  end;
end;

procedure Register;
begin
 RegisterComponents('Additional', [TAligningPanel]);
end;

constructor TAligningPanel.Create (AOwner : TComponent);
begin
 inherited Create (AOwner);
 P            := nil;
 ArraySize    := 0;
 PreviousH    := Height;
 PreviousW    := Width;
 FMoveH       := true;
 FMoveV       := true;
 FMinH        := 0;
 FMinW        := 0; 
 FSizePnlBvl  := false;
 FActive      := False; 
 FGlueLbls    := false;
 Creating     := true; 
end;

destructor TAligningPanel.Destroy;
begin
 if ArraySize > 0 then DeallocateArray;
 inherited Destroy;
end;

end.
